import configparser
import os
import sys
import typing as T
from pathlib import Path

import click

from .hasher import FileHasher

__version__ = "2019.12.4"

CONFIG_INI = "secret.ini"
CONFIG_INI_KEYS = ("hmac4csv_key", "hmac_key", "key")
CONFIG_INI_SECTIONS = ("hmac4csv", "hmac", "DEFAULT")
KEY_ENV_VAR = "HMAC4CSV_KEY"


def show_version(ctx: click.core.Context, _param: T.Any, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.echo(get_version_display())
    ctx.exit()


def get_version_display() -> str:
    return "\n".join(
        (
            f"hmac4csv {__version__} at {Path(__file__).parent.absolute()}",
            f"Python {sys.version.split(' ')[0]} at {sys.executable}",
        )
    )


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-V",
    "--version",
    is_flag=True,
    callback=show_version,
    expose_value=False,
    is_eager=True,
)
@click.argument("csvs", nargs=-1)
@click.option(
    "--key",
    "cli_key",
    default=None,
    help="String for HMAC key (e.g. --key spameggsspam)",
)
@click.option(
    "--config",
    "--ini",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="INI containing HMAC key (e.g. key=spameggsspam)",
)
@click.option("--dry-run", "-n", default=False, is_flag=True)
@click.option("--exclude", default=None, help="Comma-separated columns to exclude.")
@click.option(
    "--debug", default=False, is_flag=True, help="Print debugging; may include keys"
)
@click.pass_context
def run(
    ctx: click.core.Context,
    csvs: T.Tuple[str],
    cli_key: T.Optional[str],
    exclude: T.Optional[str],
    config: T.Optional[str],
    dry_run: bool,
    debug: bool,
) -> None:
    if debug:
        print("Called parameters:", ctx.params)

    if not csvs:
        click.echo(ctx.get_help())
        ctx.exit(1)

    if debug:
        click.echo(f"Raw CSVs in: {csvs}")
    csv_paths = get_csv_paths(csvs)
    if debug:
        click.echo(f"Parsed CSVs: {csv_paths}")

    proper_key = find_key(cli_key, config)
    if not proper_key:
        raise click.BadParameter("No key found for HMAC.")

    if debug:
        click.echo(f"Key: {proper_key}")
    exclude_list = [] if exclude is None else exclude.split(",")

    hashers = [
        FileHasher(Path(p), key=proper_key, exclude=exclude_list) for p in csv_paths
    ]

    click.echo(f"{len(hashers)} files to hash:\n")
    written = process_hashers(hashers, dry_run)
    click.echo(f"{written} hashed files written.")


def get_csv_paths(csvs: T.Sequence[str]) -> T.List[Path]:
    """Glob specification(s) of CSV paths

    Because Windows won't deal properly with globs, we'll do it internally.
    """
    nested_results = (list(Path(".").glob(s)) for s in csvs)
    return [item for sublist in nested_results for item in sublist]


def find_key(cli_key: T.Optional[str], ini_file: T.Optional[str]) -> T.Optional[str]:
    """Locate the hashing key

    Takes these in the following order:
    - CLI --key argument
    - environment variable
    - CLI-specified ini file
    - ./secret.ini
    """
    if cli_key:
        return str(cli_key)

    env_key = os.getenv(KEY_ENV_VAR)
    if env_key:
        return env_key

    ini_file = ini_file or CONFIG_INI
    if Path(ini_file).exists():
        return get_key_from_config(ini_file)

    return None


def get_key_from_config(filename: str) -> T.Optional[str]:
    """Given a config INI, retrieve a key for HMAC

    Searches for a key named in CONFIG_INI_KEYS (e.g., 'key', 'hmac_key',
    'hmac4csv_key') in either a default section or the first defined section
    and returns that value (or None).
    """
    config = configparser.ConfigParser()
    try:
        config.read(str(filename))
    except configparser.MissingSectionHeaderError:
        config.read_string("[DEFAULT]\n" + Path(filename).read_text())

    section_name = find_section_name(config.sections())

    for key_name in CONFIG_INI_KEYS:
        value = config[section_name].get(key_name)
        if value:
            return value
    return None


def find_section_name(section_list: T.List[str]) -> str:
    if not section_list:
        return "DEFAULT"
    for section_name in CONFIG_INI_SECTIONS:
        if section_name in section_list:
            return section_name
    return section_list[0]


def process_hashers(hashers: T.List[FileHasher], dry_run: bool = False) -> int:
    written = 0

    overwriting = sum(fh.dry_run() for fh in hashers)
    if dry_run:
        if overwriting:
            click.echo(f"This plan would overwrite {overwriting} existing files.")
        return written

    if overwriting:
        click.echo(f"Overwriting {overwriting} files.")
    for hasher in hashers:
        hasher.run()
        written += int(hasher.csv_out.exists())
    return written
