import csv
import hashlib
import hmac
import sys
import typing as T
from os.path import relpath
from pathlib import Path
from warnings import warn

HASH_FUNCTION = hashlib.sha256
NEWLINE = "" if sys.platform.startswith("win") else None
HERE = Path(".").resolve()

hexdigest = str


def create_in_subfolder(path: Path) -> Path:
    """Add a "hashed" subfolder to a path"""
    return path.parent / "hashed" / path.name


def simple_prefixer(path: Path) -> Path:
    """Prefix a path's filename with 'hashed_'"""
    return path.parent / ("hashed_" + path.name)


class FileHasher:

    ENCODING = "utf-8"
    FILE_ENCODING = "utf-8"

    def __init__(
        self,
        csv_in: Path,
        *,
        key: T.Union[str, bytes],
        csv_out: T.Optional[Path] = None,
        exclude: T.Optional[T.Collection[str]] = None,
        csv_out_generator: T.Callable[[Path], Path] = create_in_subfolder,
    ) -> None:
        self.hmac_key = self.format_key_as_bytes(key)
        self.exclude = set(exclude or {})
        self.csv_in = Path(csv_in).resolve()
        self.csv_out = csv_out or csv_out_generator(self.csv_in)
        self._columns: T.Sequence[str] = []

    @classmethod
    def format_key_as_bytes(cls, raw_key: T.Union[str, bytes]) -> bytes:
        if isinstance(raw_key, bytes):
            return raw_key
        if not isinstance(raw_key, str):
            raise ValueError(
                f"Invalid value {type(raw_key)} for HMAC key; "
                "should be bytes or a string to hash."
            )
        try:
            return bytes_from_hex_key(raw_key)
        except ValueError:
            pass
        encoded = raw_key.encode(encoding=cls.ENCODING)
        return HASH_FUNCTION(encoded).digest()

    def dry_run(self) -> int:
        """Print expected paths in and out, noting overwrite.

        Returns 1 for overwriting existing, 0 for a new output file."""
        exists = self.csv_out.exists()
        status = "[OVERWRITE]" if exists else ""
        print(f"  csv in:  {relpath(self.csv_in)}")
        print(f"  csv out: {relpath(self.csv_out)} {status}")
        print(f"  columns: {self.output_columns()}")
        if self.exclude & set(self.columns):
            print(f"           *not hashed")
        print("")
        return int(exists)

    def output_columns(self) -> str:
        annotated = (col + ("*" if col in self.exclude else "") for col in self.columns)
        return ", ".join(annotated)

    @property
    def columns(self) -> T.Sequence[str]:
        """Lazy loading of csv_in columns"""
        if not self._columns:
            with self.csv_in.open("r", encoding=self.FILE_ENCODING) as infile:
                reader = csv.DictReader(infile)
                self._columns = reader.fieldnames
        return self._columns

    def run(self) -> None:
        with self.csv_in.open("r", encoding=self.FILE_ENCODING) as infile:
            reader = csv.DictReader(infile)

            self.csv_out.parent.mkdir(exist_ok=True)
            with self.csv_out.open(
                "w", encoding=self.FILE_ENCODING, newline=NEWLINE
            ) as outfile:
                writer = csv.DictWriter(outfile, fieldnames=reader.fieldnames)

                writer.writeheader()
                for row in reader:
                    new_row = {k: self.calc_value(k, v) for k, v in row.items()}
                    writer.writerow(new_row)

    def calc_value(self, column: str, value: str) -> hexdigest:
        if column in self.exclude:
            return value
        return self.hash_via_hmac(self.hmac_key, value)

    @staticmethod
    def hash_via_hmac(key: bytes, message: str, encoding: str = ENCODING) -> hexdigest:
        # empty in, empty out
        if not message:
            return ""

        message_bytes = message.encode(encoding)
        hashing_object = hmac.new(key, message_bytes, HASH_FUNCTION)
        return hashing_object.hexdigest()


def bytes_from_hex_key(s: str, min_length: int = 32) -> bytes:
    if not is_hexadecimal(s):
        raise ValueError("Non-hexadecimal characters found")
    s = s[2:] if s.startswith("0x") else s
    original_bytes = bytes.fromhex(s)
    if len(original_bytes) < min_length:
        warn(
            "Warning: to use a hexademical key as-is, "
            f"it must be {min_length} bytes or greater. "
            f"Key is length {len(original_bytes)} and will be re-hashed.",
            BytesWarning,
            stacklevel=2,
        )
        raise ValueError("Hexademical key too short.")
    return original_bytes


def is_hexadecimal(s: str) -> bool:
    try:
        _ = int(s, 16)
    except (ValueError, TypeError):
        return False
    return True
