from pathlib import Path

import pytest
from click.testing import CliRunner

from hmac4csv.cli import KEY_ENV_VAR, get_csv_paths, run


@pytest.mark.parametrize(
    "incoming, outgoing",
    [
        (("*",), ["0.csv", "1.csv", "2.csv", "3.csv", "4.csv"]),
        (("*.csv",), ["0.csv", "1.csv", "2.csv", "3.csv", "4.csv"]),
        (("?.csv",), ["0.csv", "1.csv", "2.csv", "3.csv", "4.csv"]),
        (("0.csv",), ["0.csv"]),
        (("1.csv",), ["1.csv"]),
        (("0.csv", "1.csv",), ["0.csv", "1.csv"]),
        (("2.*",), ["2.csv"]),
        (("*.moo",), []),
        (("5*",), []),
        (("??.csv",), []),
    ],
)
def test_get_csv_paths(incoming, outgoing):
    runner = CliRunner()
    with runner.isolated_filesystem():
        for i in range(5):
            Path(f"{i}.csv").write_text("x")
        assert sorted(get_csv_paths(incoming)) == sorted(Path(s) for s in outgoing)


@pytest.mark.parametrize("args, exit_code", ([["-h"], 0], [["--help"], 0], [None, 1]))
def test_invoke_help(args, exit_code):
    runner = CliRunner()
    result = runner.invoke(run, args)
    assert result.exit_code == exit_code
    assert "usage" in result.output.lower()
    assert "help" in result.output.lower()


@pytest.mark.parametrize("args", (["-V"], ["--version"]))
def test_invoke_version(args):
    runner = CliRunner()
    result = runner.invoke(run, args)
    assert result.exit_code == 0
    assert "hmac4csv" in result.output.lower()
    assert "python" in result.output.lower()


def test_wet_run(run_runner):
    result, hashed = run_runner(func=run)
    assert "1 file" in result.output
    assert "1 hashed" in result.output
    assert result.exit_code == 0
    assert hashed


def test_debug(run_runner):
    key = "confuse-a-cat"
    result, hashed = run_runner(func=run, args=["--debug"], key=key)
    assert key in result.output
    assert "Raw CSVs in:" in result.output
    assert "Parsed CSVs:" in result.output
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "args, files_written", [(["-n"], 0), (["--dry-run"], 0), (None, 1),],
)
def test_dry_run_without_overwrite(run_runner, args, files_written):
    result, hashed = run_runner(func=run, args=args)
    assert result.exit_code == 0
    assert "1 file" in result.output
    assert f"{files_written} hashed files written" in result.output
    assert int(bool(hashed)) == files_written


@pytest.mark.parametrize(
    "dry_run, number_hashed", [(True, 0), (False, 1)],
)
def test_overwriting(TEST, dry_run, number_hashed):
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path(TEST.csv_name).write_text(TEST.csv_content)
        TEST.hashed_path.parent.mkdir()
        TEST.hashed_path.write_text(TEST.csv_content)
        args = [TEST.csv_name, f"--key={TEST.key}"]
        if dry_run:
            args.append("--dry-run")
        result = runner.invoke(run, args=args)
        assert "1 file" in result.output
        assert f"{number_hashed} hashed" in result.output
        assert "overwrit" in result.output.lower()
        assert result.exit_code == 0


class TestKeyInput:
    def test_bad_ini_no_key(self, run_runner):
        ini_content = "[vikings]\nmisc = spam\n"
        result, hashed = run_runner(func=run, ini_content=ini_content, key=None)
        assert result.exit_code != 0
        assert not hashed

    def test_no_key_at_all(self, run_runner, TEST):
        runner = CliRunner()
        with runner.isolated_filesystem():
            Path(TEST.csv_name).write_text(TEST.csv_content)
            result = runner.invoke(run, [TEST.csv_name])
            assert "no key" in result.output.lower()
            assert result.exit_code != 0

    def t_cli_key_simple(self, TEST):
        runner = CliRunner()
        with runner.isolated_filesystem():
            Path(TEST.csv_name).write_text(TEST.csv_content)
            result = runner.invoke(run, [TEST.csv_name, f"--key={TEST.key}"])
            assert result.exit_code == 0
            assert TEST.hashed_path.read_text().strip() == TEST.hashed_content

    def t_env_key_simple(self, TEST):
        runner = CliRunner(env={KEY_ENV_VAR: TEST.key})
        with runner.isolated_filesystem():
            Path(TEST.csv_name).write_text(TEST.csv_content)
            result = runner.invoke(run, [TEST.csv_name])
            assert result.exit_code == 0
            assert TEST.hashed_path.read_text().strip() == TEST.hashed_content

    def t_ini_key_simple(self, TEST):
        runner = CliRunner()
        with runner.isolated_filesystem():
            Path("eels.ini").write_text(f"key={TEST.key}")
            Path(TEST.csv_name).write_text(TEST.csv_content)
            result = runner.invoke(run, [TEST.csv_name, "--ini=eels.ini"])
            print(result.exit_code, result.output)
            assert result.exit_code == 0
            assert TEST.hashed_path.read_text().strip() == TEST.hashed_content

    @pytest.mark.parametrize(
        "key_name", ["key", "KEY", "HMAC_Key", "hmac4csv_key",],
    )
    @pytest.mark.parametrize(
        "ini_name, argstring",
        [
            ["secret.ini", ""],
            ["secret.ini", "--config secret.ini"],
            ["secrets.ini", "--config=secrets.ini"],
            ["arbitrary.ini", "--config arbitrary.ini"],
            ["subfolder/arbitrary.ini", "--config=subfolder/arbitrary.ini"],
        ],
    )
    def t_file_key(self, key_name, ini_name, argstring, TEST):
        runner = CliRunner()
        ini = Path(ini_name)
        with runner.isolated_filesystem():
            ini.parent.mkdir(exist_ok=True)
            ini.write_text("[hmac4csv]\nkey = spam\n")
            Path(TEST.csv_name).write_text(TEST.csv_content)
            result = runner.invoke(run, f"{TEST.csv_path.name} {argstring}")
            assert "hashed files written" in result.output
            assert result.exit_code == 0

    @pytest.mark.parametrize(
        "keyspec",
        ["key = {}", "key = {}\n", "key ={}", "key= {}", "key={}", "key={}\n",],
    )
    @pytest.mark.parametrize("prehashed_key", [True, False])
    def test_ini_parsing_key_spacing(self, run_runner, keyspec, prehashed_key, TEST):
        key = TEST.key_hashed if prehashed_key else TEST.key
        ini_content = keyspec.format(key)
        result, hashed = run_runner(func=run, ini_content=ini_content)
        assert hashed.strip() == TEST.hashed_content
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "prekey",
        [
            "[DEFAULT]\n",
            "[hmac4csv]\n",
            "[various]\n",
            "misc = spam\n",
            "[vikings]\nmisc = spam\n",
            "\n",
        ],
    )
    def test_ini_parsing_section_etc(self, run_runner, prekey, TEST):
        ini_content = prekey + f"\nkey={TEST.key}\n"
        result, hashed = run_runner(func=run, ini_content=ini_content)
        assert hashed.strip() == TEST.hashed_content
        assert result.exit_code == 0
