from base64 import b64encode
from pathlib import Path
from textwrap import dedent

import pytest

from hmac4csv.hasher import (
    HASH_FUNCTION,
    FileHasher,
    create_in_subfolder,
    simple_prefixer,
)


class TestHMACing:
    """
    Some test cases below have been borrowed to demonstrate compatibility with
    https://github.com/Coleridge-Initiative/hmac_hasher
    """

    def test_hash_via_hmac_empty_in_empty_out(self):
        assert FileHasher.hash_via_hmac(key=b"0", message="") == ""

    @pytest.mark.parametrize(
        "key, message, expected",
        [
            (
                "fakedata",
                "CalLIOpe",
                "0fd783a8226cf946b2825673ec0c2cb88e2b115c48491bbddf47f99ff510c05d",
            ),
            (
                "fakedata",
                "Calliope",
                "ffa3be8150b4c6aa6c6939411daf1419310f3851a285e27536fd2070aadd7900",
            ),
            (
                "fakedata",
                "CALLIOPE",
                "170c422bf7af219bdd28c876e50359e6fecc0234b790e25e89c59ef65b29277a",
            ),
            (
                "fakedata",
                "123456789",
                "a69ecf70cab21fdc100165faceaf87f04d0b9fb50d4dc627b04d7e5554a38bc0",
            ),
        ],
    )
    def t_from_string(self, key, message, expected):
        """Test known proper values from original script testing"""
        key = HASH_FUNCTION(key.encode(FileHasher.ENCODING)).digest()
        assert FileHasher.hash_via_hmac(key, message) == expected

    @pytest.mark.parametrize(
        "key, message, expected",
        [
            (
                b"key",
                "The quick brown fox jumps over the lazy dog",  # https://w.wiki/CpX
                "f7bc83f430538424b13298e6aa6fb143ef4d59a14946175997479dbc2d1a3cd8",
            ),
        ],
    )
    def t_from_bytes(self, key, message, expected):
        observed = FileHasher.hash_via_hmac(key, message)
        assert observed == expected


class TestFileHasher:
    @pytest.mark.parametrize("make_first", [True, False])
    @pytest.mark.parametrize("exclude", [None, ["spam"]])
    def t_dry_run(self, make_first, exclude, tmp_path, capsys):
        incoming = Path(tmp_path).with_suffix(".csv")
        incoming.write_text("spam\neggs\n")
        hasher = FileHasher(incoming, key="0", exclude=exclude)
        if make_first:
            hasher.csv_out.parent.mkdir(exist_ok=True)
            hasher.csv_out.write_text("0")
        hasher.dry_run()
        out, _err = capsys.readouterr()
        assert incoming.name in out
        assert make_first == ("overwrite" in out.lower())
        assert ("*" in out.lower()) == (exclude is not None)

    def t_end_to_end(self, tmpdir):
        incoming = dedent(
            """
            ID,SSN,FNAME,MNAME,LNAME
            555555,123456789,John,Deer,Doe
            10101010,111111111,Jane,Cat,Coe
            346712,999999999,Cynthia,Mouse,Moe
            987654,444444444,David,Fish,Foe
            23232323,888888888,Susan,Duck,Hoe
            """
        ).strip()
        expected = dedent(
            """
            ID,SSN,FNAME,MNAME,LNAME
            555555,a69ecf70cab21fdc100165faceaf87f04d0b9fb50d4dc627b04d7e5554a38bc0,82cd90dbaa364d467b126fabce3809d8bd6b7672eb4a24bb60c009ac670a6ad0,baa8f9c8a39718ff15a604e97c4f561e9a8b8c798a06e099551b445547832f50,aab80698aa7acf386ec6db017140c72830e149d1c1205c697895b5604677e5e9
            10101010,5a823abdcc524029e5497e29d65486cf69befc714758b60edf5113748afd79e3,17cb6a06a7f35616913e140be50564dde8a08d2ec780535604a308fbcba1f9fb,07c4a1f7789cc7f4bff0e566f71de7d51e02cc660a3c6215896073911f7031c7,73dac4d7317fdfbf0d18edcfb76dffd4a40c97b96c52727b16d05122bd1063f8
            346712,5f552e4659d4604b44049cc4a70d82b86c4453a0f4a88d7af2df03e8596ac4ad,bd1742af048cc0661eb94bf5fccdda8c31f8f9ca574ad969c87e1460235a622c,8a6a520479d5b8a7ae95f44702788908f71691446f30d97a8e94769ab1052584,182475eb7850b5307162a9286ad3dca789eeee6613c2932cf05ac2f21aa7e041
            987654,27becbaca10ec9cf7f2bbb4c0bece17999cad56d2d14fa5282ac2bbc5f7a82ec,1611697d4f29dfebd040b39cebfc3109cb339fbef0117a2389a64083edbe2009,38beaf392d7b2dcdd1ac8a87c46229d874f5cf3d242636b9d267c514619ea28a,74dabb3577f1f6f459f464561886c39c50872e350524e5b5bc0eee98e807a940
            23232323,cc559eb94e32af592d37a9b631a22d7ee320620d24020d59bc56b6019a593ee4,c110c28ebade4879021b2faecb323491487bacc9e196eb73377ff61c7149583e,d6d6d79b7ba72fec49c9773689d62555c3a31290384fc8b2713af4f44e63ad5e,101befa238ddee5749c2bced64e1555c0ca282109758bec5768c7e3d540fc2df
            """
        ).strip()
        incoming_csv = Path(tmpdir / "test_data.csv")
        incoming_csv.write_text(incoming)
        hasher = FileHasher(csv_in=incoming_csv, key="fakedata", exclude=["ID"])
        hasher.run()
        observed = hasher.csv_out.read_text()
        assert observed.strip() == expected.strip()


class TestKeyHandling:

    HASHED_SPAM = HASH_FUNCTION("spam".encode(FileHasher.ENCODING))
    SPAM_AS_BASE64 = b"TjiKsysQ3I28figUT1UoMK3HR4fB4sCCQDIHinnyJ/s="

    @pytest.mark.parametrize(
        "incoming",
        [
            "spam",
            HASHED_SPAM.digest(),
            HASHED_SPAM.hexdigest(),
            f"0x{HASHED_SPAM.hexdigest()}",
        ],
    )
    def t_format(self, incoming):
        assert (
            b64encode(FileHasher.format_key_as_bytes(incoming)) == self.SPAM_AS_BASE64
        )
        assert (
            b64encode(FileHasher("x.csv", key=incoming).hmac_key) == self.SPAM_AS_BASE64
        )

    @pytest.mark.parametrize("badvalue", [0, -1, 53511, ("moo", "meow"), {3, 5}])
    def t_key_format_errors(self, badvalue):
        with pytest.raises(ValueError):
            FileHasher.format_key_as_bytes(badvalue)

    @pytest.mark.parametrize(
        "shortkey", ["535110", "4920646f6e27742062656c69657665206974"],
    )
    def t_warn_and_hash_short_hex(self, shortkey):
        with pytest.warns(Warning):
            hasher = FileHasher(csv_in="n/a", key=shortkey)
            assert hasher.hmac_key != bytes.fromhex(shortkey)


class TestCSVOutGenerators:
    def t_csv_file_handling(self, tmp_path):
        incoming = Path(tmp_path)
        hasher = FileHasher(csv_in=incoming, key="fakedata")
        assert hasher.csv_in.name == incoming.name == hasher.csv_out.name
        assert hasher.csv_in.parent == hasher.csv_out.parent.parent

    @pytest.mark.parametrize(
        "inpath, outpath",
        [
            ["tmp/spam.csv/", "tmp/hashed/spam.csv"],
            ["c:/eggs.csv", "c:/hashed/eggs.csv"],
        ],
    )
    def test_create_in_subfolder(self, inpath, outpath):
        assert create_in_subfolder(Path(inpath)) == Path(outpath)

    @pytest.mark.parametrize(
        "inpath, outpath",
        [
            ["tmp/spam.csv/", "tmp/hashed_spam.csv"],
            ["c:/eggs.csv", "c:/hashed_eggs.csv"],
        ],
    )
    def test_simple_prefixer(self, inpath, outpath):
        assert simple_prefixer(Path(inpath)) == Path(outpath)
